import { CROSS_AXIS_SIZE } from "../consts";
import { getScreenVisibleRectangle } from "../hitTest";
import { Vec4 } from "../math/Vec4";
import { intersection, isInside } from "../math/utils";
import { BaseView } from "./BaseView";
import { UserEventType } from "./eventTypes";
import { Overflow } from "./styling";
/**
 * `BaseView` but with event listeners.
 */
export class View extends BaseView {
    props;
    _eventListeners = [];
    /**
     * Controlled by `EventManager`. Needed for dispatching mouseEnter and mouseLeave events.
     */
    _isMouseOver = false;
    _scrolling;
    constructor(props) {
        super(props);
        this.props = props;
        this._scrolling = {
            xActive: false,
            xHovered: false,
            yActive: false,
            yHovered: false,
        };
        this.onScroll = this.onScroll.bind(this);
        this.handleMouseDownScrollStart = this.handleMouseDownScrollStart.bind(this);
        this.handleMouseMoveScrollHovering = this.handleMouseMoveScrollHovering.bind(this);
        this.onMouseEnter = this.onMouseEnter.bind(this);
        this.onMouseLeave = this.onMouseLeave.bind(this);
        // For mouse-interacting with the scrollbar.
        // TODO: this is done when creating the node but scrollbar can be added later (like with
        // Overflow.Auto). What then?
        if (this._style.overflowX === Overflow.Scroll || this._style.overflowY === Overflow.Scroll) {
            this._eventListeners.push([UserEventType.Scroll, this.onScroll], [UserEventType.MouseDown, this.handleMouseDownScrollStart], [UserEventType.MouseMove, this.handleMouseMoveScrollHovering], [UserEventType.MouseEnter, this.onMouseEnter], [UserEventType.MouseLeave, this.onMouseLeave]);
        }
        if (props.onClick) {
            this._eventListeners.push([UserEventType.MouseClick, props.onClick]);
        }
        if (props.onMouseEnter) {
            this._eventListeners.push([UserEventType.MouseEnter, props.onMouseEnter]);
        }
        if (props.onMouseLeave) {
            this._eventListeners.push([UserEventType.MouseLeave, props.onMouseLeave]);
        }
    }
    onScroll(event) {
        this._state.scrollX = Math.min(Math.max(this._state.scrollX + Math.round(event.delta.x), 0), this._state.scrollWidth - this._state.clientWidth);
        this._state.scrollY = Math.min(Math.max(this._state.scrollY + Math.round(event.delta.y), 0), this._state.scrollHeight - this._state.clientHeight);
    }
    onMouseEnter(_) {
        // No-op but important to keep this._isMouseOver up to date.
    }
    onMouseLeave(_) {
        this._scrolling.xHovered = false;
        this._scrolling.yHovered = false;
    }
    handleMouseDownScrollStart(_) {
        if (this._scrolling.xHovered) {
            this._scrolling.xActive = true;
        }
        if (this._scrolling.yHovered) {
            this._scrolling.yActive = true;
        }
    }
    handleMouseMoveScrollHovering(event) {
        if (this._scrolling.xActive || this._scrolling.yActive) {
            return;
        }
        if (this._isMouseOver) {
            // Screen space rectangle of the node's visible area, including scrollbars.
            const rectangle = getScreenVisibleRectangle(this);
            if (this._state.hasHorizontalScrollbar) {
                const scrollbar = new Vec4(rectangle.x, rectangle.y + rectangle.w - CROSS_AXIS_SIZE, this._state.clientWidth, CROSS_AXIS_SIZE);
                const clippedScrollbar = intersection(rectangle, scrollbar);
                this._scrolling.xHovered = isInside(event.position, clippedScrollbar);
            }
            if (this._state.hasVerticalScrollbar) {
                const scrollbar = new Vec4(rectangle.x + rectangle.z - CROSS_AXIS_SIZE, rectangle.y, CROSS_AXIS_SIZE, this._state.clientHeight);
                const clippedScrollbar = intersection(rectangle, scrollbar);
                this._scrolling.yHovered = isInside(event.position, clippedScrollbar);
            }
        }
        else {
            this._scrolling.xHovered = false;
            this._scrolling.yHovered = false;
        }
    }
}
//# sourceMappingURL=View.js.map